/*
 * Copyright (C) 2019  Kevin Whitaker <eyecreate@eyecreate.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "miflora.h"
#include <QDataStream>

void MiFlora::startSearch()
{
    qDebug() << "Starting search...";
    agent = new QBluetoothDeviceDiscoveryAgent();
    agent->setLowEnergyDiscoveryTimeout(5000);
    devices.clear();
    connect(agent, &QBluetoothDeviceDiscoveryAgent::deviceDiscovered, this, &MiFlora::foundDevice);
    agent->start(QBluetoothDeviceDiscoveryAgent::LowEnergyMethod);
}

void MiFlora::stopSearch()
{
    if(agent != NULL) { 
        qDebug() << "Stopping Search.";
        agent->stop();
        agent->disconnect();
        delete agent;
    }
}

void MiFlora::updateDataFromDevice ( QString mac )
{
    for(BluetoothDevices *info: devices) {
        if(info->device().address().toString() == mac) {
            qDebug() << "Discovering on:" + info->device().name();
            currentController = QLowEnergyController::createCentral(info->device());
            currentController->setRemoteAddressType(QLowEnergyController::PublicAddress);
            connect(currentController, &QLowEnergyController::serviceDiscovered, this, &MiFlora::serviceDiscovered);
            connect(currentController, static_cast<void (QLowEnergyController::*)(QLowEnergyController::Error)>(&QLowEnergyController::error), this, &MiFlora::logControllerError);
            connect(currentController, &QLowEnergyController::connected, this, [this]{
                currentController->discoverServices();
            });
            currentController->connectToDevice();
        }
    }
}

QAbstractTableModel * MiFlora::getModel()
{
    return floraModel;
}

void MiFlora::foundDevice ( const QBluetoothDeviceInfo& device )
{
    if(device.coreConfigurations() & QBluetoothDeviceInfo::LowEnergyCoreConfiguration && device.address().toString().startsWith("C4:7C")) {
        devices.append(new BluetoothDevices(device));
        qDebug() << "Added device";
        emit newDeviceFound();
    }
}

QQmlListProperty<BluetoothDevices> MiFlora::getDeviceList()
{
    return QQmlListProperty<BluetoothDevices>(this, devices);
}

void MiFlora::logControllerError ( QLowEnergyController::Error err )
{
    qDebug() << "Error:" << err;
    emit errorHappened(currentController->errorString());
}

void MiFlora::logServiceError(QLowEnergyService::ServiceError err)
{
    qDebug() << "Service Error:" << err;
    emit errorHappened("Possible Read/Write error!");
}


void MiFlora::sensorServiceStateChanges ( QLowEnergyService::ServiceState state )
{
    if(state == QLowEnergyService::ServiceState::ServiceDiscovered) {
        for(QLowEnergyCharacteristic chrt : sensorService->characteristics()) {
            if(chrt.uuid().toUInt16() == magicChar) {
                sensorService->writeCharacteristic(chrt, magicBytes);
            }
        }
    }
}

void MiFlora::historyServiceStateChanges ( QLowEnergyService::ServiceState state )
{
    if(state == QLowEnergyService::ServiceState::ServiceDiscovered) {
        for(QLowEnergyCharacteristic chrt : historyService->characteristics()) {
            if(chrt.uuid().toUInt16() == historyTimeChar) {
                qDebug() << "Asking for device's view on time.";
                historyService->readCharacteristic(chrt);
            }
        }
    }
}

void MiFlora::serviceCharWritten(QLowEnergyCharacteristic changedChar, QByteArray value)
{
    if(changedChar.uuid().toUInt16() == magicChar) {
        //Used to only check sensor data after writting magic bytes.
        for(QLowEnergyCharacteristic chrt : sensorService->characteristics()) {
            if(chrt.uuid().toUInt16() == batteryFirmwareChar) {
                sensorService->readCharacteristic(chrt);
            }
            if(chrt.uuid().toUInt16() == sensorsChar) {
                sensorService->readCharacteristic(chrt);
            }
        }
    }
}

void MiFlora::serviceCharRead(QLowEnergyCharacteristic readChar, QByteArray value)
{
    if(readChar.uuid().toUInt16() == batteryFirmwareChar) {
        QDataStream parser(value);
        parser >> battery;
        parser.skipRawData(1);
        QByteArray version_buf(5, Qt::Uninitialized);
        parser.readRawData(version_buf.data(), 5);
        version = QString(version_buf);
        emit batteryChanged(battery);
        qDebug() << "Firmware: " << version;
    } else if(readChar.uuid().toUInt16() == sensorsChar) {
        quint16 origTemp;
        QDataStream parser(value);
        parser.setByteOrder(QDataStream::LittleEndian);
        parser >> origTemp;
        temp = origTemp/ 10.0; //original value in 0.1 C
        parser.skipRawData(1);
        parser >> bright;
        parser >> moisture;
        parser >> conduct;
        emit brightnessChanged(bright);
        emit temperatureChanged(temp);
        emit moistureChanged(moisture);
        emit conductionChanged(conduct);
    }
}

QLowEnergyCharacteristic MiFlora::getCharFromValue(QLowEnergyService *service, quint16 characteristic)
{
    for(QLowEnergyCharacteristic chrt : service->characteristics()) {
        if(chrt.uuid().toUInt16() == characteristic) {
            return chrt;
        }
    }
    return QLowEnergyCharacteristic();
}

void MiFlora::historyServiceCharRead(QLowEnergyCharacteristic readChar, QByteArray value)
{
    if(readChar.uuid().toUInt16() == historyReaderChar) {
        QDataStream parser(value);
        parser.setByteOrder(QDataStream::LittleEndian);
        quint16 size;
        parser >> size;
        //change connect signals/slots to perform history read task.
        disconnect(historyService, &QLowEnergyService::characteristicRead, this, &MiFlora::historyServiceCharRead);
        connect(historyService, &QLowEnergyService::characteristicRead, this, &MiFlora::historyServiceCharReadData);
        disconnect(historyService, &QLowEnergyService::characteristicWritten, this, &MiFlora::historyServiceCharWritten);
        connect(historyService, &QLowEnergyService::characteristicWritten, this, &MiFlora::historyServiceCharWrittenData);
        //Take size given and start by giving the last 48 numbers(if that many) to represent the last day of data. TODO: look to have amount of time be configurable.
        lastHistoryEntry = 0;
        if(size > 48) {
            currentHistoryEntry = 48;
        } else {
            currentHistoryEntry = size;
        }
        floraModel->clear();
        //Ask for first history value
        QByteArray historyIndex(QByteArray::fromHex("a1"));
        QDataStream historyParser(&historyIndex, QIODevice::ReadWrite);
        historyParser.setByteOrder(QDataStream::LittleEndian);
        historyParser.skipRawData(1);
        historyParser << currentHistoryEntry;
        historyService->writeCharacteristic(getCharFromValue(historyService, historyControllerChar), historyIndex);
    } else if(readChar.uuid().toUInt16() == historyTimeChar) {
        //Determine when device started counting by comparing system and device time.
        quint32 time;
        QDataStream parser(value);
        parser.setByteOrder(QDataStream::LittleEndian);
        parser >> time;
        qint64 systemTime = QDateTime::currentSecsSinceEpoch();
        deviceStartTime = systemTime - time;
        //Now we start the history reading process.
        qDebug() << "Writing history init.";
        historyService->writeCharacteristic(getCharFromValue(historyService, historyControllerChar), historyReadInit);
    }
}

void MiFlora::historyServiceCharReadData(QLowEnergyCharacteristic readChar, QByteArray value)
{
    //We are in a read loop. Read data and determine if we should stop asking for history.
    if(value == QByteArray::fromHex("ffffffffffffffffffffffffffffffff") || value == QByteArray::fromHex("00000000000000000000000000000000") || value == QByteArray::fromHex("aabbccddeeff99887766554433221110")) {
        qDebug() << "invalid history response:" << value;
    } else {
        quint16 origTemp;
        quint32 time;
        quint32 bright;
        quint8 moisture;
        quint16 conduct;
        float temp;
        QDataStream parser(value);
        parser.setByteOrder(QDataStream::LittleEndian);
        parser >> time;
        parser >> origTemp;
        temp = origTemp/ 10.0; //original value in 0.1 C
        parser.skipRawData(1);
        parser >> bright;
        parser >> moisture;
        parser >> conduct;
        QDateTime convTime;
        convTime.setSecsSinceEpoch(deviceStartTime+time);
        //Write out items to table
        floraModel->addData(convTime, temp, bright, moisture, conduct);
    }
    
    if(lastHistoryEntry == currentHistoryEntry) {
        disconnect(historyService, &QLowEnergyService::characteristicRead, this, &MiFlora::historyServiceCharReadData);
        connect(historyService, &QLowEnergyService::characteristicRead, this, &MiFlora::historyServiceCharRead);
        disconnect(historyService, &QLowEnergyService::characteristicWritten, this, &MiFlora::historyServiceCharWrittenData);
        connect(historyService, &QLowEnergyService::characteristicWritten, this, &MiFlora::historyServiceCharWritten);
    } else {
        //Get next history entry
        currentHistoryEntry -= 1;
        QByteArray historyIndex(QByteArray::fromHex("a1"));
        QDataStream parser(&historyIndex, QIODevice::ReadWrite);
        parser.setByteOrder(QDataStream::LittleEndian);
        parser.skipRawData(1);
        parser << currentHistoryEntry;
        historyService->writeCharacteristic(getCharFromValue(historyService, historyControllerChar), historyIndex);
    }
}

void MiFlora::historyServiceCharWrittenData(QLowEnergyCharacteristic changedChar, QByteArray value)
{
    //We are in a read loop. Ask to read from history what should have just been requested by a write.
    historyService->readCharacteristic(getCharFromValue(historyService, historyReaderChar));
}


void MiFlora::historyServiceCharWritten(QLowEnergyCharacteristic changedChar, QByteArray value)
{
    if(changedChar.uuid().toUInt16() == historyControllerChar) {
        for(QLowEnergyCharacteristic chrt : historyService->characteristics()) {
            if(chrt.uuid().toUInt16() == historyReaderChar) {
                historyService->readCharacteristic(chrt);
            }
        }
    }
}

void MiFlora::serviceDiscovered ( const QBluetoothUuid& gatt )
{
    if(gatt == QBluetoothUuid(dataServiceChar)) {
        sensorService = currentController->createServiceObject(gatt);
        connect(sensorService, &QLowEnergyService::stateChanged, this, &MiFlora::sensorServiceStateChanges);
        connect(sensorService, static_cast<void (QLowEnergyService::*)(QLowEnergyService::ServiceError)>(&QLowEnergyService::error), this, &MiFlora::logServiceError);
        connect(sensorService, &QLowEnergyService::characteristicWritten, this, &MiFlora::serviceCharWritten);
        connect(sensorService, &QLowEnergyService::characteristicRead, this, &MiFlora::serviceCharRead);
        sensorService->discoverDetails();
    } else if(gatt == QBluetoothUuid(historyServiceChar)) {
        historyService = currentController->createServiceObject(gatt);
        connect(historyService, &QLowEnergyService::stateChanged, this, &MiFlora::historyServiceStateChanges);
        connect(historyService, static_cast<void (QLowEnergyService::*)(QLowEnergyService::ServiceError)>(&QLowEnergyService::error), this, &MiFlora::logServiceError);
        connect(historyService, &QLowEnergyService::characteristicWritten, this, &MiFlora::historyServiceCharWritten);
        connect(historyService, &QLowEnergyService::characteristicRead, this, &MiFlora::historyServiceCharRead);
        historyService->discoverDetails();
    }
}



